# pristine-lfs
#
# store pristine tarballs in Git LFS
#
# Copyright (C) 2019—2021 Collabora Ltd
# Copyright (C) 2019—2021 Andrej Shadura <andrew.shadura@collabora.co.uk>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import argparse
import logging
import os
import sys
from fnmatch import fnmatch
from gettext import gettext as _
from pathlib import Path
from typing import (
    IO,
    Iterable,
    Optional,
    Sequence,
    Union
)

import sh
from debian import deb822
from debian.changelog import Changelog, Version

from .errors import (
    CommandFailed,
    DifferentFilesExist,
    GitError,
    UnsupportedHashAlgorithm,
)
from .log import logger
from .util import (
    check_branch,
    checkout_lfs_file,
    checkout_package,
    commit_lfs_file,
    commit_lfs_files,
    find_branch,
    find_remote_branches,
    list_git_files,
    track_remote_branch,
    verify_lfs_file,
)


def do_commit(tarball: IO[bytes], branch: str, message: Optional[str] = None, force_overwrite: bool = False, **kwargs):
    """
    Commit the open file to a branch using Git LFS.
    Set force_overwrite to overwrite an existing file with the same name and a different checksum.
    Message may contain "%s" which gets replaced with a comma-separate list of the file committed.
    """
    if check_branch(branch) is None:
        if find_remote_branches(branch):
            track_remote_branch(branch)
    commit_lfs_file(tarball, branch, message, overwrite=force_overwrite)


def do_commit_files(tarballs: Sequence[IO[bytes]], branch: str, message: Optional[str] = None,
                    force_overwrite: bool = False, **kwargs):
    """
    Commit open files to a branch using Git LFS.
    Set force_overwrite to overwrite existing files with same names and different checksums.
    Message may contain "%s" which gets replaced with a comma-separate list of the file committed.
    """
    if check_branch(branch) is None:
        if find_remote_branches(branch):
            track_remote_branch(branch)
    commit_lfs_files(tarballs, branch, message, overwrite=force_overwrite)


def do_checkout(branch: str, tarball: Optional[str] = None, outdir: Union[str, Path] = '.', full: bool = False,
                package: Optional[str] = None, version: Union[str, Version, None] = None, **kwargs):
    """
    Check out one or multiple files.
    If tarball is non-None:
        * file name only: tarball to check out to outdir.
        * path with to file: the location where to check out to
    If tarball is None:
        * if package and version are specified, that version
          is checked out
        * a tarball corresponding to the latest entry in
          debian/changelog is found and checked out

    When full is set, tarball is a name of a .dsc file;
    all corresponding tarball *and* the .dsc file are
    checked out. If None, debian/changelog is used
    to find the set of files to check out.
    """
    branch = find_branch(branch)
    if tarball:
        path, slash, tarball = tarball.rpartition('/')
        if path:
            outdir = path
    else:
        if not package or not version:
            changelog = Path("debian/changelog")
            with changelog.open() as f:
                ch = Changelog(f, max_blocks=1)
            package, version = ch.package, ch.version

    outdir = Path(outdir)
    outdir.mkdir(parents=True, exist_ok=True)

    if full:
        if tarball:
            dsc_file = tarball
        else:
            fver = Version(version)
            fver.epoch = None
            dsc_file = f'{package}_{fver}.dsc'
        logger.info(_("Checking out file {} in {}").format(dsc_file, outdir))
        checkout_lfs_file(branch, dsc_file, outdir)
        if dsc_file.endswith('.dsc'):
            with (outdir / dsc_file).open('r') as dsc:
                d = deb822.Dsc(dsc)
            package = str(d['Source'])
            version = Version(d['Version'])
            files = [f["name"] for f in d["Files"]]
            checkout_package(package, version, branch, outdir, files)
    else:
        if tarball:
            logger.info(_("Checking out file {} in {}").format(tarball, outdir))
            checkout_lfs_file(branch, tarball, outdir)
        else:
            checkout_package(str(package), Version(version), branch, outdir)


def do_list(branch: str, **kwargs) -> Iterable[str]:
    """
    List all files on the specified branch except .gitattributes.
    """
    branch = find_branch(branch)
    for f in list_git_files(branch):
        if f != '.gitattributes':
            yield f


def do_import(dsc: IO[str], branch: str, message: Optional[str] = None, force_overwrite: bool = False,
              full: bool = False, **kwargs):
    """
    Import all tarballs and detached signatures related to an open .dsc file.
    Set force_overwrite to overwrite an existing file with the same name and a different checksum.
    Message may contain "%s" which gets replaced with a comma-separate list of the file committed.
    When full is set, the .dsc file and the Debian packaging tarball are also imported.
    """
    d = deb822.Dsc(dsc)
    package = d['Source']
    version = Version(d['Version'])

    tarball_glob = f'{package}_{version.upstream_version}.orig.tar.*'
    component_tarball_glob = f'{package}_{version.upstream_version}.orig-*.tar.*'
    dsc_dir = os.path.dirname(dsc.name)

    if check_branch(branch) is None:
        if find_remote_branches(branch):
            track_remote_branch(branch)

    tarballs = [os.path.join(dsc_dir, f['name']) for f in d['Files']
                if full or fnmatch(f['name'], tarball_glob) or fnmatch(f['name'], component_tarball_glob)]
    if full:
        tarballs += [dsc.name]

    if tarballs:
        logger.info("Importing: %s" % " ".join(tarballs))
        commit_lfs_files([open(tarball, 'rb') for tarball in tarballs], branch, message, overwrite=force_overwrite)


def do_verify(branch: str, tarball: Union[str, Path], **kwargs) -> bool:
    """
    Verify an existing tarball is the same as the one committed with pristine-lfs.
    """
    branch = find_branch(branch)
    tarball = Path(tarball)
    return verify_lfs_file(branch, tarball)


def main(*args):
    prog = os.path.basename(sys.argv[0])

    parser = argparse.ArgumentParser(description=_('store pristine tarballs in Git LFS'), prog=prog)
    if args and hasattr(parser, 'exit_on_error'):
        parser.exit_on_error = False

    parser.add_argument('-v', '--verbose', action='count', help=_('be more verbose'))
    parser.add_argument('--debug', action='store_const', const=2, dest='verbose', help=_('be debuggingly verbose'))
    parser.set_defaults(verbose=0, func=lambda *x, **kw: parser.print_usage(file=sys.stderr))
    subparsers = parser.add_subparsers(required=False)

    parser_commit = subparsers.add_parser('commit', help=_('commit a tarball'))
    parser_commit.add_argument('--force-overwrite', action='store_true', help=_('overwrite already stored files'))
    parser_commit.add_argument('-m', '--message', default=None, help=_('commit message'))
    parser_commit.add_argument('-b', '--branch', default='pristine-lfs', help=_('branch to store metadata on'))
    parser_commit.add_argument('tarball', type=argparse.FileType('rb'), help=_('tarball to commit'))
    parser_commit.add_argument('upstream', nargs='?', default=None, help=_('ignored'))
    parser_commit.set_defaults(func=do_commit)

    # we have to do some trickery since argparse doesn’t support this syntax natively
    parser_checkout = subparsers.add_parser('checkout', help=_('checkout a tarball'))
    parser_checkout.add_argument('-b', '--branch', default='pristine-lfs', help=_('branch to store metadata on'))
    parser_checkout.add_argument('--full', default=False, action='store_true',
                                 help=_('also check out all related files of the Debian package'))
    parser_checkout.add_argument('-o', '--outdir', default='.', help=_('output directory for the tarball'))
    checkout_group = parser_checkout.add_mutually_exclusive_group(required=True)
    checkout_group.add_argument('--auto', default=False, action='store_true',
                                help=_('check out all tarballs required by the currently checked out Debian package'))
    checkout_group.add_argument('tarball', nargs='?', default=None, help=_('tarball to check out'))
    parser_checkout.set_defaults(func=do_checkout)

    parser_list = subparsers.add_parser('list', help=_('list tarballs stored in the repository'))
    parser_list.add_argument('-b', '--branch', default='pristine-lfs', help=_('branch to store metadata on'))
    parser_list.set_defaults(func=do_list)

    parser_import = subparsers.add_parser('import-dsc', help=_('import tarballs and their signatures from a .dsc'))
    parser_import.add_argument('dsc', type=argparse.FileType('r'), help='.dsc file to use')
    parser_import.add_argument('--force-overwrite', action='store_true', help=_('overwrite already stored files'))
    parser_import.add_argument('--full', default=False, action='store_true',
                               help=_('also import all related files of the Debian package'))
    parser_import.add_argument('-m', '--message', default=None, help=_('commit message'))
    parser_import.add_argument('-b', '--branch', default='pristine-lfs', help=_('branch to store metadata on'))
    parser_import.set_defaults(func=do_import)

    parser_verify = subparsers.add_parser('verify', help=_('verify a tarball against Git LFS metadata'))
    parser_verify.add_argument('-b', '--branch', default='pristine-lfs', help=_('branch to store metadata on'))
    parser_verify.add_argument('tarball', help=_('tarball to verify'))
    parser_verify.set_defaults(func=do_verify)

    args = parser.parse_args(args or sys.argv[1:])

    logging.basicConfig(format='{levelname[0]}: {message!s}', style='{', level=(logging.WARNING - 10 * args.verbose))

    # sh is printing debug on the info channel
    logging.getLogger(sh.__name__).setLevel(logging.WARNING - 10 * (args.verbose - 1))

    try:
        ret = args.func(**vars(args))
        if isinstance(ret, Iterable):
            for item in ret:
                print(item)
        elif isinstance(ret, bool):
            return 0 if ret else 1
    except CommandFailed as e:
        print(_('Failed to run %s:') % e.full_cmd, file=sys.stderr)
        print(e.stderr.decode(sh.DEFAULT_ENCODING, "replace"), file=sys.stderr)
        return e.exit_code
    except FileNotFoundError as e:
        print(_('abort: file %s not found ') % e.filename, file=sys.stderr)
        return os.EX_OSERR
    except OSError as e:
        print(_('I/O error: %s') % e, file=sys.stderr)
        return os.EX_OSERR
    except KeyboardInterrupt:
        print(file=sys.stderr)
        print(_('about: Interrupted by user'), file=sys.stderr)
        return 1
    except (DifferentFilesExist, GitError, UnsupportedHashAlgorithm) as e:
        print(_("abort: %s\n") % e, file=sys.stderr)
        return 1



# Extremely simplified type stubs for git and git.lfs.
# Extend as needed.
#
# Copyright (C) 2021 Collabora Ltd
# Copyright (C) 2021 Andrej Shadura <andrew.shadura@collabora.co.uk>
#
# SPDX-License-Identifier: GPL-2.0-or-later

class git:
    class lfs:
        def __init__(*args): ...
        def __call__(self, *args, **kwargs): ...
        def smudge(self, *args, **kwargs): ...
        def clean(self, *args, **kwargs): ...
    def __init__(*args, **kwargs): ...
    def __call__(self, *args, **kwargs): ...



import hashlib
import os
import tarfile
from pathlib import Path
from textwrap import dedent
from typing import NamedTuple, Optional

import pytest


try:
    import sh
except ImportError:
    assert False, "These tests require working python-sh"

try:
    from sh.contrib import git
except ImportError:
    assert False, "These tests require Git"


class WorkDir(NamedTuple):
    path: Path
    tarball: Optional[Path] = None
    size: Optional[int] = None
    sha: Optional[str] = None


@pytest.fixture
def empty_git_repo(tmp_path):
    os.chdir(tmp_path)
    git.init()
    git.config('user.name', 'Committer Name')
    git.config('user.email', 'name@example.org')
    git.config('commit.gpgsign', 'false')

    assert (tmp_path / '.git').is_dir()
    yield WorkDir(tmp_path)


@pytest.fixture(params=['/bin/true'])
def fake_tarball(empty_git_repo, request):
    repo, *_ = empty_git_repo

    tarball = Path('true_0.orig.tar.gz')
    with tarfile.open(tarball, mode='w:gz') as f:
        f.add(request.param)

    size = tarball.stat().st_size
    h = hashlib.sha256()
    with open(tarball, mode='rb') as f:
        while True:
            chunk = f.read(h.block_size)
            if not chunk:
                break
            h.update(chunk)
    sha = h.hexdigest()

    yield WorkDir(repo, tarball, size, sha)


@pytest.fixture
def fake_pristine_lfs(fake_tarball):
    repo, tarball, size, sha = fake_tarball

    git.checkout('-b', 'pristine-lfs')
    git.lfs.track('*.tar.*')
    f = Path('.gitattributes')
    assert f.is_file()
    assert f.stat().st_size != 0
    git.add(f)

    git.add(tarball)
    git.commit(message=f'add {tarball.name}')
    commit = git('rev-parse', 'pristine-lfs^{tree}').strip('\n')
    pointer = git('cat-file', 'blob', f"{commit}:{tarball.name}")
    assert pointer == dedent(
        f"""
        version https://git-lfs.github.com/spec/v1
        oid sha256:{sha}
        size {size}
        """).lstrip('\n'), 'Object pointer doesn’t match the object'

    yield WorkDir(repo, tarball, size, sha)


@pytest.fixture
def test_git_repo(fake_pristine_lfs):
    repo, tarball, size, sha = fake_pristine_lfs

    # clean index
    empty = git('hash-object', '-t', 'tree', '/dev/null').strip('\n')
    git('read-tree', empty)
    # start a new empty branch
    git.checkout(orphan='debian/test')
    changelog = Path('debian/changelog')
    changelog.parent.mkdir()
    changelog.write_text(dedent(
        """
        true (0-0) UNRELEASED; urgency=medium
        
            * Empty.
        
            -- Nobody <nobody@example.org>  Thu, 01 Jan 1970 12:00:00 +0100
        """).lstrip('\n'))  # noqa: W293
    git.add(changelog)
    git.commit(message='pretend packaging')
    git.reset(hard=True)
    git.clean(force=True)

    yield WorkDir(repo, tarball, size, sha)


def pytest_collection_modifyitems(items):
    # sort doctests before normal tests
    # but run smoke tests first
    def test_key(item):
        return type(item).__name__, not [mark.name == 'smoke' for mark in item.iter_markers()]
    items.sort(key=test_key)



# Extremely simplified type stubs for the Git wrapper.
# Extend as needed.
#
# Copyright (C) 2021 Collabora Ltd
# Copyright (C) 2021 Andrej Shadura <andrew.shadura@collabora.co.uk>

from sh.contrib import git as sh_git


class Git(sh_git):
    lfs: sh_git.lfs


git: Git



# pristine-lfs
#
# errors for pristine-lfs
#
# Copyright (C) 2021 Collabora Ltd
# Copyright (C) 2021 Andrej Shadura <andrew.shadura@collabora.co.uk>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from gettext import gettext as _

from sh import ErrorReturnCode as CommandFailed  # noqa: F401


class DifferentFilesExist(Exception):
    files: list[str]

    def __init__(self, files: list[str]):
        self.files = files

    def __str__(self):
        return _("would overwrite files: {files}").format(files=', '.join(self.files))


class UnsupportedHashAlgorithm(Exception):
    algo: str

    def __init__(self, algo: str):
        self.algo = algo

    def __str__(self):
        return _("unsupported hash algorithm {algo}").format(
            algo=self.algo,
        )


class GitError(Exception):
    pass


class GitFileNotFound(GitError):
    filename: str
    branch: str

    def __init__(self, filename: str, branch: str):
        self.filename = filename
        self.branch = branch

    def __str__(self):
        return _('{filename} not found on branch {branch}').format(
            filename=self.filename,
            branch=self.branch,
        )


class GitBranchNotFound(GitError):
    branch: str

    def __init__(self, branch: str):
        self.branch = branch

    def __str__(self):
        return _('No branch {branch} found, not even among remote branches').format(
            branch=self.branch,
        )



# Wrapper for Git and Git LFS
#
# Copyright (C) 2021 Collabora Ltd
# Copyright (C) 2021 Andrej Shadura <andrew.shadura@collabora.co.uk>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import os
from pathlib import Path
from typing import Optional

from sh.contrib import git as _git


def git(*args, index: Optional[Path] = None, **kwargs):
    """
    Wrapper for Git and Git LFS

    Differences from sh.contrib.git:
        * accepts index argument to set GIT_INDEX_FILE
    """
    e = os.environ.copy()
    if index:
        e["GIT_INDEX_FILE"] = str(index)
    return _git(*args, **kwargs, _env=e)


git.lfs = _git.lfs



# pristine-lfs
#
# logging
#
# Copyright (C) 2021 Collabora Ltd
# Copyright (C) 2021 Andrej Shadura <andrew.shadura@collabora.co.uk>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging


logger = logging.getLogger("pristine-lfs")



from textwrap import dedent

from pristine_lfs import do_import, do_list


def test_pristine_lfs_import_dsc(fake_tarball):
    repo, tarball, size, sha = fake_tarball

    base, *_ = tarball.name.split('.')
    package, version = base.split('_')

    outdir = repo / 'tmp'
    outdir.mkdir()
    dsc = outdir / f'{package}_{version}.dsc'
    dsc.write_text(dedent(
        f"""
        Format: 3.0 (quilt)
        Source: {package}
        Binary: {package}
        Architecture: any
        Version: {version}
        Package-List:
         {package} deb utils optional arch=any
        Checksums-Sha1:
         0000000000000000000000000000000000000000 0 {tarball.name}
        Checksums-Sha256:
         0000000000000000000000000000000000000000000000000000000000000000 0 {tarball.name}
        Files:
         00000000000000000000000000000000 0 {tarball.name}
        """).lstrip('\n'))

    tarball.rename(outdir / tarball.name)

    do_import(dsc.open(), 'pristine-lfs', f'import {package}')
    do_import(dsc.open(), 'pristine-lfs', f'import {package}')
    do_import(dsc.open(), 'pristine-lfs-source', f'import {package}', full=True)

    ret = list(do_list(branch='pristine-lfs'))
    assert ['true_0.orig.tar.gz'] == ret

    ret = list(do_list(branch='pristine-lfs-source'))
    assert ['true_0.dsc', 'true_0.orig.tar.gz'] == ret



from pristine_lfs import do_list


def test_pristine_lfs_list(test_git_repo):
    ret = list(do_list(branch='pristine-lfs'))
    assert [test_git_repo.tarball.name] == ret, 'Expected tarball not found'

    ret = list(do_list(branch='debian/test'))
    assert ["debian/changelog"] == ret, 'This branch should only have the changelog'



from pristine_lfs import do_commit, do_verify


def test_pristine_lfs_commit(fake_tarball):
    repo, tarball, size, sha = fake_tarball

    do_commit(tarball.open('rb'), branch='pristine-lfs')
    do_commit(tarball.open('rb'), branch='pristine-lfs', message='blip %s %s %s')
    do_commit(tarball.open('rb'), branch='pristine-lfs', message='blip')

    # verify the file has indeed been committed
    assert do_verify(branch='pristine-lfs', tarball=tarball), 'Tarball doesn’t match the one in Git LFS'

    # now try a different file
    tarball.write_text('dummy')
    assert not do_verify(branch='pristine-lfs', tarball=tarball), 'Corrupted tarball not detected'



# pristine-lfs
#
# Git and Git LFS routines
# This requires Git and git-lfs to be installed.
#
# Copyright (C) 2019—2021 Collabora Ltd
# Copyright (C) 2019—2021 Andrej Shadura <andrew.shadura@collabora.co.uk>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import hashlib
import os
from contextlib import contextmanager
from fnmatch import fnmatch, fnmatchcase
from gettext import gettext as _
from pathlib import Path
from typing import IO, Any, Generator, Iterable, Mapping, Optional, Sequence, Tuple, Union, cast

import sh
from debian.changelog import Version

from .errors import (
    DifferentFilesExist,
    GitBranchNotFound,
    GitFileNotFound,
    UnsupportedHashAlgorithm,
)
from .gitwrap import git
from .log import logger


gitattributes = """*.tar.* filter=lfs diff=lfs merge=lfs -text
*.tar.*.asc -filter -diff merge=binary -text
*.dsc -filter !diff merge=binary !text
"""

pre_push_hook = """#!/bin/sh -e

case "$GIT_LFS_SKIP_PUSH" in
    true | on | 1)
        exit 0
        ;;
    *)
        git lfs pre-push "$@"
        ;;
esac
"""

supported_lfs_hashsums = ['sha256']


RETURN_CMD = {} if sh.__version__.startswith('1.') else {'_return_cmd': True}


@contextmanager
def open_index(name: str) -> Generator[Path, None, None]:
    """Context manager for a named temporary Git index file."""
    index = Path(git_dir()) / f"index-{name}"
    if index.exists():
        index.unlink()
    try:
        yield index
    finally:
        if index.exists():
            index.unlink()


AttributeValue = Union[str, bool, None]
Attribute = Tuple[str, AttributeValue]


def parse_attr(attr: str) -> Attribute:
    """Parse a Git attribute into its value.

    >>> parse_attr('attr')
    ('attr', True)
    >>> parse_attr('-attr')
    ('attr', False)
    >>> parse_attr('!attr')
    ('attr', None)
    >>> parse_attr('attr=text')
    ('attr', 'text')
    """
    if attr.startswith('!'):
        return attr[1:], None
    if attr.startswith('-'):
        return attr[1:], False
    if '=' not in attr:
        return attr, True
    return cast(Attribute, tuple(attr.split('=', maxsplit=1)))


def parse_git_attributes(s: str) -> Iterable[tuple[str, Mapping[str, AttributeValue]]]:
    """Parse Git attributes from a string.

    >>> list(parse_git_attributes('''
    ... ab*     merge=filfre
    ... abc     -foo -bar
    ... *.c     frotz
    ... *
    ... '''))
    [('ab*', {'merge': 'filfre'}),
     ('abc', {'foo': False, 'bar': False}),
     ('*.c', {'frotz': True}),
     ('*', {})]
    >>> list(parse_git_attributes('[attr]binary -diff -merge -text'))
    [('[attr]binary', {'diff': False, 'merge': False, 'text': False})]
    >>> list(parse_git_attributes('''
    ... # this is a comment
    ... a*      foo !bar -baz
    ... abc     foo bar baz
    ... *.jpg -text -diff
    ... *               text=auto
    ... *.txt           text
    ... *.vcproj        text eol=crlf
    ... *.sh            text eol=lf
    ... *.jpg           -text
    ... '''))
    [('a*', {'foo': True, 'bar': None, 'baz': False}),
     ('abc', {'foo': True, 'bar': True, 'baz': True}),
     ('*.jpg', {'text': False, 'diff': False}),
     ('*', {'text': 'auto'}),
     ('*.txt', {'text': True}),
     ('*.vcproj', {'text': True, 'eol': 'crlf'}),
     ('*.sh', {'text': True, 'eol': 'lf'}),
     ('*.jpg', {'text': False})]
    """
    lines = [line.strip() for line in s.splitlines()]
    # strip comments, split each line by consecutive spaces
    lines = [line.split() for line in lines if len(line) and not line.startswith('#')]
    for line in lines:
        if len(line):
            glob, *attrs = line
            yield glob, {k: v for k, v in [
                parse_attr(a) for a in attrs
            ]}


default_gitattributes = list(parse_git_attributes(gitattributes))


def check_branch(name: str) -> Optional[str]:
    """Check a branch exists, return the hash it points at, if it does.

    None if there’s no such branch.
    """
    try:
        return git('show-ref', '--heads', '--hash', '--', name)
    except sh.ErrorReturnCode:
        return None


def git_dir() -> str:
    """Return the path to the `.git` directory."""
    return git('rev-parse', '--git-dir').strip('\n')


def git_head() -> str:
    """Return the name of the current branch, or the SHA1 of the current commit if we're in a detached HEAD state."""
    return git('rev-parse', '-q', '--verify', '--symbolic-full-name', 'HEAD', _ok_code=[0, 1]).strip('\n')


def branch_remote(name: str) -> Optional[str]:
    """Find an upstream remote for a tracking branch."""
    if name.startswith('refs/heads/'):
        name = name[len("refs/heads/"):]

    return git('config', 'branch.%s.remote' % name, ok_code=[0, 1]).strip('\n') or None


def find_remote_branches(name: str) -> list[tuple[str, str]]:
    """Find remote branches with a given name.

    Return a list of tuples, each tuple containing the SHA1 and the full name of a remote branch.
    """
    try:
        branches = [line.split(' ') for line in git('show-ref', '--', name).splitlines()]
        return [(b[0], b[1]) for b in branches if b[1].startswith('refs/remotes/')]
    except sh.ErrorReturnCode:
        return []


def preferred_remote_branch(remote_branches: list[tuple[str, str]]) -> tuple[str, str]:
    """Choose one remote branch out of a list.

    Return the first remote branch that matches the current remote, or the first remote
    branch if there is no current remote. The current remote is a remote for the currently
    checked out branch.
    """
    logger.debug("Remote branches: %r", remote_branches)
    current_remote = branch_remote(git_head())
    logger.debug("Current remote: %r", current_remote)
    remote_branches = [
        (commit, ref) for (commit, ref) in remote_branches
        if current_remote and ref.startswith('refs/remotes/' + current_remote)
    ] + remote_branches
    return remote_branches[0]


def track_remote_branch(name: str):
    """Find and track a remote branch for a given branch.

    Find all remote branches with a given name, and then create a local branch with the
    same name that tracks the remote branch picked by preferred_remote_branch.
    """
    remote_branches = find_remote_branches(name)
    if len(remote_branches) == 0:
        raise RuntimeError('remote branch expected but not found')
    commit, branch = preferred_remote_branch(remote_branches)
    git('branch', '--track', name, branch)


def find_branch(branch: str) -> str:
    """Find a branch.

    Return one of the following:
     * an existing branch of the given name
     * a remote branch picked by preferred_remote_branch

    Raise an exception if all fails.
    """
    if check_branch(branch) is None:
        remote_branches = find_remote_branches(branch)
        if remote_branches:
            commit, branch = preferred_remote_branch(remote_branches)
        else:
            raise GitBranchNotFound(branch)
    return branch


def store_lfs_object(io: Any) -> str:
    """Store a file-like object in Git LFS, and return the Git LFS pointer for the stored blob."""
    return str(git.lfs.clean(io.name, _in=io))


def store_git_object(io: Any) -> str:
    """Store a file-like object in Git, and return its hash."""
    return git('hash-object', '-w', '--stdin', _in=io).strip('\n')


def stage_file(filename: Union[str, bytes], io: Any, index: Optional[Path] = None):
    """Take a file, store it in the Git object store, and then add it to the index."""
    blob = store_git_object(io)
    if isinstance(filename, bytes):
        filename = filename.decode()
    git('update-index', '--add', '--replace', '--cacheinfo', "100644,%s,%s" % (blob, filename), index=index)


def create_commit(branch: str, message: str, index: Optional[Path] = None) -> str:
    """Create a commit with the given message and index, and update the given branch to point to it."""
    tree = git('write-tree', index=index).strip('\n')
    if not len(tree):
        raise RuntimeError('write-tree failed')

    if check_branch(branch) is not None:
        commit = git('commit-tree', tree, '-p', branch, _in=message).strip('\n')
    else:
        commit = git('commit-tree', tree, _in=message).strip('\n')
    if not len(commit):
        raise RuntimeError('commit-tree failed')

    git('update-ref', 'refs/heads/%s' % branch, commit)

    return commit


def refresh_main_index():
    """Refresh the main index, ignoring submodules, missing files and unmerged files."""
    git('update-index', '--ignore-submodules', '-q', '--ignore-missing', '--unmerged', '--refresh')


def parse_diff_entry(entry: str) -> Mapping[str, str]:
    r"""
    Parse an entry in git-diff-tree format into a dictionary.

    >>> parse_diff_entry(
    ... ":100644 100644 777f41fb18215a23a779a831b38d24ff6171775a c5ac094096d35c687c0cc5054b5a1433c293ebd5 M\tpristine_lfs/main.py"
    ... )
    {'srcmode': '100644',
     'dstmode': '100644',
     'srchash': '777f41fb18215a23a779a831b38d24ff6171775a',
     'dsthash': 'c5ac094096d35c687c0cc5054b5a1433c293ebd5',
     'srcname': 'pristine_lfs/main.py',
     'dstname': '',
     'status': 'M'}
    >>> parse_diff_entry(
    ... ":100644 100644 f0adb461cd94419245cf0b16a0edf1e2daf6688b f0adb461cd94419245cf0b16a0edf1e2daf6688b R100\told\tnew"
    ... )
    {'srcmode': '100644',
     'dstmode': '100644',
     'srchash': 'f0adb461cd94419245cf0b16a0edf1e2daf6688b',
     'dsthash': 'f0adb461cd94419245cf0b16a0edf1e2daf6688b',
     'srcname': 'old',
     'dstname': 'new',
     'status': 'R100'}
    """
    treediff, names = entry.split('\t', maxsplit=1)
    srcmode, dstmode, srchash, dsthash, status = treediff.lstrip(':').split(' ')
    srcname, _, dstname = names.partition('\t')
    return {
        'srcmode': srcmode,
        'dstmode': dstmode,
        'srchash': srchash,
        'dsthash': dsthash,
        'srcname': srcname,
        'dstname': dstname,
        'status': status,
    }


def commit_lfs_file(io: IO[bytes], branch: str, template: Optional[str] = None, overwrite: bool = False):
    """Store the file in the LFS storage and commit it to a branch."""
    commit_lfs_files([io], branch, template=template, overwrite=overwrite)


def commit_lfs_files(ios: Sequence[IO[bytes]], branch: str, template: Optional[str] = None, overwrite: bool = False):
    """Store the files in the LFS storage and commit them to a branch."""
    # make sure the pre-push hook has been set up
    hook_path = Path(git_dir()) / 'hooks' / 'pre-push'
    if not hook_path.is_file():
        try:
            hook_path.parent.mkdir(exist_ok=True)
            hook_path.write_text(pre_push_hook)
            hook_path.chmod(0o755)
        except IOError as e:
            logger.warning(_('Failed to set up pre-push hook: %s') % e.strerror)

    with open_index("pristine-lfs") as index:
        # make sure we include all previously committed files
        if check_branch(branch) is not None:
            git('update-index', '--index-info', index=index, _in=git('ls-tree', '-r', '--full-name', branch))

        # make sure .gitattributes is present
        stage_file('.gitattributes', gitattributes, index=index)

        for io in ios:
            filename = os.path.basename(io.name)
            if not is_lfs_managed(os.path.basename(filename), default_gitattributes):
                stage_file(filename, io, index=index)
            else:
                metadata = store_lfs_object(io)
                stage_file(filename, metadata, index=index)

        if check_branch(branch) is not None:
            diff = git('diff-index', '--cached', branch, index=index).strip().splitlines()
            if not diff:
                logger.info(_("Nothing to commit"))
                return
            parsed_diff = [parse_diff_entry(d) for d in diff]
            overwritten = [d['srcname'] for d in parsed_diff if d['srchash'] != ('0' * 40) and d['srcname'] != '.gitattributes']
            if any(overwritten) and not overwrite:
                raise DifferentFilesExist(overwritten)

        if not template:
            template = "pristine-lfs data for %s"

        message = template.replace('%s', ', '.join([os.path.basename(io.name) for io in ios]))

        create_commit(branch, message, index=index)

        refresh_main_index()


def list_lfs_files(branch: str) -> list[str]:
    """Return a list of all the files tracked by Git LFS in the given branch."""
    return git.lfs('ls-files', '--name-only', branch).splitlines()


def parse_entry(entry: str) -> tuple[str, ...]:
    info, name = entry.split('\t')
    mode, type, hash = info.split(' ')
    return mode, type, hash, name


def list_git_files(branch: str) -> Mapping[str, str]:
    """Return a dictionary mapping file names to SHA1 hashes for all files in the given branch."""
    entries = [parse_entry(line) for line in git('ls-tree', '-r', '--full-name', branch).splitlines()]
    return {e[3]: e[2] for e in entries if e[1] == 'blob'}


def is_lfs_managed(filename: str, attributes: Iterable[tuple[str, Mapping[str, AttributeValue]]]):
    """Tell if the file is managed by Git LFS.

    If the filename matches a pattern in the attributes, and the filter attribute is set
    to lfs, then the file is LFS-managed.
    """
    lfs_managed = False
    for pattern, attrs in attributes:
        if fnmatchcase(filename, pattern):
            if 'filter' in attrs:
                lfs_managed = attrs['filter'] == 'lfs'
    return lfs_managed


def checkout_lfs_file(branch: str, filename: str, outdir: Union[str, Path] = '.'):
    """
    Check out a file from a Git branch:
     * if it's managed by LFS, use Git LFS
     * otherwise check it out normally
    """
    files = list_git_files(branch)
    if '.gitattributes' in files:
        attributes = parse_git_attributes(git('cat-file', 'blob', files['.gitattributes']))
    else:
        attributes = []

    if filename not in files:
        raise GitFileNotFound(filename, branch)

    with (Path(outdir) / filename).open(mode='wb') as tarball:
        if is_lfs_managed(filename, attributes):
            metadata = git('cat-file', 'blob', files[filename])
            git.lfs.smudge(filename, _out=tarball, _in=metadata)
        else:
            git('cat-file', 'blob', files[filename], _out=tarball)


def verify_lfs_file(branch: str, tarball: Path) -> bool:
    """Check if the tarball's hash matches the one stored in the Git repo."""
    files = list_git_files(branch)

    filename = tarball.name
    if filename not in files:
        raise GitFileNotFound(filename, branch)

    metadata = git('cat-file', 'blob', files[filename], **RETURN_CMD)
    parsed_metadata = dict(parse_pointer(metadata))
    oid = parsed_metadata['oid']
    algo, hashsum = oid.split(':', 1)
    if algo not in supported_lfs_hashsums:
        raise UnsupportedHashAlgorithm(algo)

    h = getattr(hashlib, algo)()
    with open(tarball, mode='rb') as f:
        while True:
            chunk = f.read(h.block_size)
            if not chunk:
                break
            h.update(chunk)
    calc_hashsum = h.hexdigest()
    if hashsum == calc_hashsum:
        logger.info(f"{algo} hash for the tarball: {hashsum}, matches the stored one")
    else:
        logger.warning(_("%(tarball)s does not match stored hash (expected %(stored_hash)s, got %(tarball_hash)s)") % {
            'tarball': filename,
            'stored_hash': hashsum,
            'tarball_hash': calc_hashsum,
        })
    return hashsum == calc_hashsum


def checkout_package(package: str, version: Version, branch: str, outdir: Union[str, Path],
                     requested: Optional[Sequence[str]] = None):
    """Check out all files necessary for a given version of a given package."""
    logger.info(_("Checking out files for {package} version {version} to {outdir}:").format(
        package=package,
        version=version, outdir=outdir
    ))
    tarball_glob = f'{package}_{version.upstream_version}.orig.tar.*'
    component_tarball_glob = f'{package}_{version.upstream_version}.orig-*.tar.*'

    files = list_git_files(branch)
    if requested:
        # TODO: handle missing files
        tarballs = [f for f in files if f in requested]
    else:
        tarballs = [f for f in files if fnmatch(f, tarball_glob) or fnmatch(f, component_tarball_glob)]

    for f in tarballs:
        logger.info("         ... {}".format(f))
        checkout_lfs_file(branch, f, outdir)
    logger.info(_("Done."))


def parse_pointer(pointer: IO[str]) -> Iterable[tuple[str, str]]:
    """
    Parse Git LFS file pointer into a mapping of keys to values.

    >>> from io import StringIO
    >>> dict(parse_pointer(StringIO('''version https://git-lfs.github.com/spec/v1
    ... oid sha256:4d7a214614ab2935c943f9e0ff69d22eadbb8f32b1258daaa5e2ca24d17e2393
    ... size 12345
    ... ''')))
    {'version': 'https://git-lfs.github.com/spec/v1',
     'oid': 'sha256:4d7a214614ab2935c943f9e0ff69d22eadbb8f32b1258daaa5e2ca24d17e2393',
     'size': '12345'}
    """
    for line in pointer:
        key, value = line.rstrip('\n').split(' ', 1)
        yield key, value



#!/usr/bin/env python3

import fnmatch
import os
from setuptools import setup, distutils


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def glob(fname):
    return fnmatch.filter(os.listdir(os.path.abspath(os.path.dirname(__file__))), fname)

def generate_manpage(src, dst):
    import docutils.core
    distutils.log.info("generating a manpage from %s to %s", src, dst)
    docutils.core.publish_file(source_path=src, destination_path=dst, writer_name='manpage')

def man_name(fname):
    import re
    matches = re.compile(r'^:Manual section: *([0-9]*)', re.MULTILINE).search(read(fname))
    if matches:
        section = matches.groups()[0]
    else:
        section = '7'
    base = os.path.splitext(fname)[0]
    manfname = base + '.' + section
    return manfname

def man_path(fname):
    category = fname.rsplit('.', 1)[1]
    return os.path.join('share', 'man', 'man' + category), [fname]

def man_files(pattern):
    return list(map(man_path, map(man_name, glob(pattern))))

__manpages__ = 'pristine-lfs.rst'

from setuptools.command import build_py


build_py_org = build_py.build_py
class build_py_new(build_py_org):
    def run(self):
        build_py_org.run(self)
        if not self.dry_run:
            for page in glob(__manpages__):
                generate_manpage(page, man_name(page))
build_py.build_py = build_py_new

__name__ = "pristine-lfs"

setup(
    data_files = [
        (os.path.join('share', 'doc', __name__), glob('*.rst')),
        (os.path.join('share', 'doc', __name__), ['CONTRIBUTORS', 'COPYING'])
    ] + man_files(__manpages__)
)



# Type hints for the subset of sh we use.
# Extend as needed.
#
# Copyright (C) 2021 Collabora Ltd
# Copyright (C) 2021 Andrej Shadura <andrew.shadura@collabora.co.uk>
#
# SPDX-License-Identifier: GPL-2.0-or-later

class ErrorReturnCode(Exception): ...

DEFAULT_ENCODING: str
__version__: str



from pristine_lfs import do_checkout
from pristine_lfs.util import Version


def test_pristine_lfs_simple_checkout(fake_pristine_lfs):
    repo, tarball, size, sha = fake_pristine_lfs
    outdir = repo / 'tmp'

    do_checkout('pristine-lfs', tarball=tarball.name, outdir=outdir)
    assert len(list(outdir.glob('**/'))) == 1, list(outdir.glob('**/'))
    assert (outdir / tarball.name).is_file(), 'Extracted tarball not found'
    assert (outdir / tarball.name).stat().st_size == size, 'Extracted tarball of a wrong size'


def test_pristine_lfs_auto_checkout(test_git_repo):
    repo, tarball, size, sha = test_git_repo
    outdir = repo / 'tmp-explicit'

    do_checkout('pristine-lfs', tarball=tarball.name, outdir=outdir)
    assert len(list(outdir.glob('*'))) == 1, list(outdir.glob('*'))
    assert (outdir / tarball.name).is_file()
    assert (outdir / tarball.name).stat().st_size == size

    outdir = repo / 'tmp-auto'

    do_checkout('pristine-lfs', tarball=None, outdir=outdir)
    assert len(list(outdir.glob('*'))) == 1, list(outdir.glob('*'))
    assert (outdir / tarball.name).is_file(), 'Extracted tarball not found'
    assert (outdir / tarball.name).stat().st_size == size, 'Extracted tarball of a wrong size'

    outdir = repo / 'tmp-empty'

    do_checkout('pristine-lfs', package='true', version='1', outdir=outdir)
    assert len(list(outdir.glob('*'))) == 0, list(outdir.glob('*'))
    assert not (outdir / tarball.name).is_file(), 'Found a tarball which should not be there'

    do_checkout('pristine-lfs', package='true', version=Version('1'), outdir=outdir)
    assert len(list(outdir.glob('*'))) == 0, list(outdir.glob('*'))
    assert not (outdir / tarball.name).is_file(), 'Found a tarball which should not be there'

    do_checkout('pristine-lfs', package='true', version='0', outdir=outdir)
    assert len(list(outdir.glob('*'))) == 1, list(outdir.glob('*'))
    assert (outdir / tarball.name).is_file(), 'Extracted tarball not found'

    outdir = repo / 'tmp-empty-2'

    do_checkout('pristine-lfs', package='true', version='0', outdir=outdir)
    assert len(list(outdir.glob('*'))) == 1, list(outdir.glob('*'))
    assert (outdir / tarball.name).is_file(), 'Extracted tarball not found'



from textwrap import dedent

import pytest
from sh.contrib import git

from pristine_lfs import do_commit, do_commit_files
from pristine_lfs.errors import DifferentFilesExist


def test_pristine_lfs_commit(fake_tarball):
    repo, tarball, size, sha = fake_tarball

    do_commit(tarball.open('rb'), branch='pristine-lfs')
    do_commit(tarball.open('rb'), branch='pristine-lfs', message='blip %s %s %s')
    do_commit(tarball.open('rb'), branch='pristine-lfs', message='blip')

    # verify the file has indeed been committed
    commit = git('rev-parse', 'pristine-lfs^{tree}').strip('\n')
    pointer = git('cat-file', 'blob', f"{commit}:{tarball.name}")
    assert pointer == dedent(
        f"""
        version https://git-lfs.github.com/spec/v1
        oid sha256:{sha}
        size {size}
        """).lstrip('\n'), 'Object pointer doesn’t match the object'

    stored = repo / '.git' / 'lfs' / 'objects' / sha[:2] / sha[2:4] / sha
    assert stored.is_file(), 'Object has not been stored by LFS'


def test_pristine_lfs_commit_overwrite(fake_tarball):
    repo, tarball, size, sha = fake_tarball

    do_commit(tarball.open('rb'), branch='pristine-lfs')
    tarball.write_text('Text')

    with pytest.raises(DifferentFilesExist):
        do_commit(tarball.open('rb'), branch='pristine-lfs', message='blip %s %s %s')


def test_pristine_lfs_commit_overwrite_forced(fake_tarball):
    repo, tarball, size, sha = fake_tarball

    do_commit(tarball.open('rb'), branch='pristine-lfs')
    tarball.write_text('Text')

    do_commit(tarball.open('rb'), branch='pristine-lfs', message='blip %s %s %s', force_overwrite=True)


def test_pristine_lfs_commit_files(fake_tarball):
    repo, tarball, size, sha = fake_tarball

    sig = tarball.with_suffix('.gz.asc')
    fake_sig_text = 'Fake signature'
    sig.write_text(fake_sig_text)

    do_commit_files([tarball.open('rb'), sig.open('rb')], branch='pristine-lfs', message='blip %s %s %s')

    commit = git('rev-parse', 'pristine-lfs^{tree}').strip('\n')
    pointer = git('cat-file', 'blob', f"{commit}:{tarball.name}")
    assert pointer == dedent(
        f"""
        version https://git-lfs.github.com/spec/v1
        oid sha256:{sha}
        size {size}
        """).lstrip('\n'), 'Object pointer doesn’t match the object'

    sig_text = git('cat-file', 'blob', f"{commit}:{sig.name}")
    assert sig_text == fake_sig_text
